@@ -1,10 +1,15 @@
+from __future__ import annotations
+
 import re
 from collections import defaultdict
 from ctypes import c_float
 from sys import stderr
-from typing import Any, Callable, Dict, Iterable, Tuple, Union
+from typing import Any, Callable, Iterable, Union
 
-from .errors import MoreThanOneAltAllele, NotExactlyOneValue
+import numpy as np
+
+from .errors import MoreThanOneAltAlleleError, NotExactlyOneValueError
+from .sequence_ontology import Consequences, Term
 
 
 def float32(val: str) -> float:
@@ -12,41 +17,57 @@ def float32(val: str) -> float:
 
 
 # If NoValue inherits from str, re.search("something", NoValue()) does not error
-# but just comes up empty handed, which is convenient behaviour.
+# but just comes up empty-handed, which is convenient behaviour.
 # This way, we do not have to special case / monkey patch / wrap the regex module.
 class NoValue(str):
-    def __lt__(self, other):
+    warnings: set[str] = set()
+
+    def __lt__(self, other) -> bool:
         return False
 
-    def __gt__(self, other):
+    def __gt__(self, other) -> bool:
         return False
 
-    def __le__(self, other):
+    def __le__(self, other) -> bool:
         return False
 
-    def __ge__(self, other):
+    def __ge__(self, other) -> bool:
         return False
 
-    def __eq__(self, other):
+    def __eq__(self, other) -> bool:
         return False
 
-    def __ne__(self, other):
+    def __ne__(self, other) -> bool:
         return True
 
-    def __bool__(self):
+    def __bool__(self) -> bool:
         return False
 
-    def __str__(self):
+    def __str__(self) -> str:
         # nonexistent fields will result in an empty string
         # should be configurable in upcoming versions
         return ""
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"{self.__class__.__name__}()"
 
     def __hash__(self) -> int:
         return super().__hash__()
 
+    def __getattr__(self, item):
+        if item not in self.warnings:
+            self.warnings.add(item)
+            print(
+                f"Warning: Trying to access non-existent attribute '{item}' of NoValue."
+                " Returning NA instead.\n"
+                "This warning will only be printed once per attribute.\n"
+                "It either indicates a typo in the attribute name or "
+                "the access of an attribute of a field with a custom type "
+                "which is empty/has no value.",
+                file=stderr,
+            )
+        return self
+
 
 NA = NoValue()
 
@@ -54,7 +75,7 @@ NA = NoValue()
 class InfoTuple:
     """A container that lazily evaluates None to NA in case of access."""
 
-    def __init__(self, values):
+    def __init__(self, values) -> None:
         self.values = values
 
     def __getitem__(self, spec):
@@ -65,13 +86,13 @@ class InfoTuple:
             return NA
         return values
 
-    def __str__(self):
+    def __str__(self) -> str:
         return self.values.__str__()
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"{self.__class__.__name__}(values={self.values!r})"
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self.values)
 
     def __eq__(self, other):
@@ -89,7 +110,7 @@ class InfoTuple:
                     f" (value type: {type(other.values)})"
                     if isinstance(other, InfoTuple)
                     else ""
-                )
+                ),
             )
 
     def __hash__(self):
@@ -97,25 +118,46 @@ class InfoTuple:
 
 
 IntFloatStr = Union[int, float, str]
+NvIntFloatStr = Union[IntFloatStr, NoValue]
+NvInt = Union[int, NoValue]
+NvFloat = Union[float, NoValue]
 
 
-def type_info(value, number=".", field=None, record_idx=None):
+def type_info(
+    value,
+    number=".",
+    field=None,
+    record_idx=None,
+) -> NvIntFloatStr | InfoTuple:
     if value is None:
         return NA
+    if value is NA:
+        if number == "0":
+            return False
+        return NA
     if number == "0":
-        return NA if value is None else value
+        return False if value is NA else value
+    if isinstance(value, np.ndarray):
+        value = tuple(value.tolist())
+    elif isinstance(value, list):
+        value = tuple(value)
     if number == "A":
-        if len(value) != 1:
-            raise MoreThanOneAltAllele()
-        return value[0] if value[0] is not None else NA
+        if isinstance(value, tuple):
+            if len(value) > 1:
+                raise MoreThanOneAltAlleleError()
+            return value[0] if value[0] is not None else NA
+        return value
     if number == "R":
-        if len(value) != 2:
-            raise MoreThanOneAltAllele()
-        return InfoTuple(value)
+        if len(value) > 2:
+            raise MoreThanOneAltAlleleError()
+        if len(value) == 1:
+            return InfoTuple(tuple(value) + (None,))
+        else:
+            return InfoTuple(tuple(value))
     if number == "1":
         if isinstance(value, tuple):
             if len(value) != 1:
-                raise NotExactlyOneValue(field, len(value), record_idx)
+                raise NotExactlyOneValueError(field, len(value), record_idx)
             return value[0] if value[0] is not None else NA
         return value if value is not None else NA
     if isinstance(value, tuple):
@@ -125,37 +167,42 @@ def type_info(value, number=".", field=None, record_idx=None):
 
 
 class PosRange:
-    def __init__(self, start: int, end: int):
+    def __init__(self, start: NvInt, end: NvInt, raw: str) -> None:
         self.start = start
         self.end = end
+        self._raw = raw
+
+    raw = property(lambda self: self._raw, None, None)
 
     @classmethod
-    def from_snpeff_str(cls, value: str) -> "PosRange":
-        pos, length = [int(v.strip()) for v in value.split("/")]
-        return cls(pos, pos + length)
+    def from_snpeff_str(cls, value: str) -> PosRange:
+        pos, length = (int(v.strip()) for v in value.split("/"))
+        return cls(pos, pos + length, value)
 
     @classmethod
-    def from_vep_str(cls, value: str) -> "PosRange":
-        start, end = (
-            #  the "-" is optional, so vep either has either start and end position
-            [NA if v == "?" else int(v) for v in map(str.strip, value.split("-"))]
-            if "-" in value
-            #  or start position only
-            else [int(value.strip())] * 2
-        )
-        return cls(start, end)
+    def from_vep_str(cls, value: str) -> PosRange:
+        if "-" in value:
+            s, e = map(str.strip, value.split("-", 1))
+            start: NvInt = NA if s == "?" else int(s)
+            end: NvInt = NA if e == "?" else int(e)
+            return cls(start, end, value)
+        elif len(value) > 0:
+            start = int(value.strip())
+            return cls(start, start + 1, value)
+        else:
+            return cls(NA, NA, value)
 
     @property
     def length(self):
-        if self.end != NA and self.start != NA:
+        if self.end is not NA and self.start is not NA:
             return self.end - self.start
         else:
             return NA
 
-    def __str__(self):
+    def __str__(self) -> str:
         return f"(start: {self.start}, end: {self.end}, length: {self.length})"
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return f"{self.__class__.__name__}(start={self.start!r}, end={self.end!r})"
 
 
@@ -163,7 +210,7 @@ def na_func() -> NoValue:
     return NA
 
 
-AnnotationType = Union[IntFloatStr, Iterable["IntFloatStr"], NoValue]
+AnnotationType = Union[IntFloatStr, Iterable[IntFloatStr], NoValue]
 
 
 class AnnotationEntry:
@@ -172,8 +219,8 @@ class AnnotationEntry:
         name: str,
         typefunc: Callable[[str], Any] = str,
         nafunc: Callable[[], Any] = na_func,
-        description: str = None,
-    ):
+        description: str | None = None,
+    ) -> None:
         self._name = name
         self._typefunc = typefunc
         self._nafunc = nafunc
@@ -183,7 +230,7 @@ class AnnotationEntry:
     def name(self):
         return self._name
 
-    def convert(self, value: str) -> Tuple[str, Any]:
+    def convert(self, value: str) -> Any:
         if value:
             return self._typefunc(value)
         else:
@@ -192,7 +239,7 @@ class AnnotationEntry:
     def description(self):
         return self._description
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return (
             f"{self.__class__.__name__}("
             f"name={self._name!r}, "
@@ -208,10 +255,10 @@ class AnnotationListEntry(AnnotationEntry):
         self,
         name: str,
         sep: str,
-        typefunc: Callable[[str], Any] = None,
+        typefunc: Callable[[str], Any] | None = None,
         inner_typefunc: Callable[[str], Any] = lambda x: x,
         **kwargs,
-    ):
+    ) -> None:
         typefunc = (
             typefunc
             if typefunc
@@ -220,10 +267,39 @@ class AnnotationListEntry(AnnotationEntry):
         super().__init__(name, typefunc, nafunc=lambda: [], **kwargs)
 
 
+class AnnotationSetEntry(AnnotationEntry):
+    def __init__(
+        self,
+        name: str,
+        sep: str,
+        typefunc: Callable[[str], Any] | None = None,
+        inner_typefunc: Callable[[str], Any] = lambda x: x,
+        nafunc: Callable[[], Any] = lambda: set(),
+        **kwargs,
+    ) -> None:
+        typefunc = (
+            typefunc
+            if typefunc
+            else lambda v: {inner_typefunc(x.strip()) for x in v.split(sep)}
+        )
+        super().__init__(name, typefunc, nafunc=nafunc, **kwargs)
+
+
+class ConsequenceEntry(AnnotationSetEntry):
+    def __init__(self, name: str, **kwargs) -> None:
+        super().__init__(
+            name,
+            sep="&",
+            typefunc=lambda v: Consequences(Term(t.strip()) for t in v.split("&")),
+            nafunc=lambda: Consequences(),
+            **kwargs,
+        )
+
+
 class AnnotationListDictEntry(AnnotationEntry):
-    def __init__(self, name: str, nafunc=lambda: None, **kwargs):
+    def __init__(self, name: str, nafunc=lambda: None, **kwargs) -> None:
         def typefunc(x):
-            key_value_tuples = map(lambda v: v.split(":", maxsplit=1), x.split("&"))
+            key_value_tuples = (v.split(":", maxsplit=1) for v in x.split("&"))
             mapping = defaultdict(list)
             for key, value in key_value_tuples:
                 mapping[key].append(value)
@@ -236,26 +312,29 @@ class AnnotationListDictEntry(AnnotationEntry):
         super().__init__(name=name, typefunc=typefunc, **kwargs)
 
 
-class RangeTotal(object):
-    def __init__(self, r, total):
+class RangeTotal:
+    def __init__(self, r: range, total: int, raw: str) -> None:
         self.range = r
         self.total = total
+        self._raw = raw
+
+    raw = property(lambda self: self._raw, None, None)
 
     @classmethod
-    def from_vep_string(cls, value: str) -> "RangeTotal":
+    def from_vep_string(cls, value: str) -> RangeTotal:
         v = value.split("/")
         r = [int(s) for s in v[0].split("-")]
         if len(r) == 1:
-            return cls(range(r[0], r[0] + 1), int(v[1]))
+            return cls(range(r[0], r[0] + 1), int(v[1]), value)
         elif len(r) == 2:
-            return cls(range(r[0], r[1] + 1), int(v[1]))
+            return cls(range(r[0], r[1] + 1), int(v[1]), value)
         else:
             raise ValueError(
                 "Found more than two values separated by '-', "
-                "expected only a single int, or two ints separated by '-'."
+                "expected only a single int, or two ints separated by '-'.",
             )
 
-    def __str__(self):
+    def __str__(self) -> str:
         if len(self.range) == 1:
             return f"number / total: {self.range.start} / {self.total}"
         else:
@@ -264,7 +343,7 @@ class RangeTotal(object):
                 f"{self.range.start} - {self.range.stop - 1} / {self.total}"
             )
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return (
             f"{self.__class__.__name__}"
             f"(range=range({self.range.start!r}, {self.range.stop!r}), "
@@ -273,31 +352,34 @@ class RangeTotal(object):
 
 
 class AnnotationRangeTotalEntry(AnnotationEntry):
-    def __init__(self, name: str, **kwargs):
+    def __init__(self, name: str, **kwargs) -> None:
         super().__init__(name, typefunc=RangeTotal.from_vep_string, **kwargs)
 
 
-class NumberTotal(object):
-    def __init__(self, number, total):
+class NumberTotal:
+    def __init__(self, number: int, total: int, raw: str) -> None:
         self.number = number
         self.total = total
+        self._raw = raw
+
+    raw = property(lambda self: self._raw, None, None)
 
     @classmethod
-    def from_vep_string(cls, value: str) -> "NumberTotal":
+    def from_vep_string(cls, value: str) -> NumberTotal:
         v = value.split("/")
-        return cls(int(v[0]), int(v[1]))
+        return cls(int(v[0]), int(v[1]), value)
 
-    def __str__(self):
+    def __str__(self) -> str:
         return f"number / total: {self.number} / {self.total}"
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return (
             f"{self.__class__.__name__}(number={self.number!r}, total={self.total!r})"
         )
 
 
 class AnnotationNumberTotalEntry(AnnotationEntry):
-    def __init__(self, name: str, **kwargs):
+    def __init__(self, name: str, **kwargs) -> None:
         super().__init__(name, typefunc=NumberTotal.from_vep_string, **kwargs)
 
 
@@ -305,21 +387,21 @@ PREDICTION_SCORE_REGEXP: re.Pattern = re.compile(r"(.*)\((.*)\)")
 
 
 class AnnotationPredictionScoreEntry(AnnotationEntry):
-    def __init__(self, name: str, **kwargs):
-        def typefunc(x: str) -> Dict[str, float]:
+    def __init__(self, name: str, **kwargs) -> None:
+        def typefunc(x: str) -> dict[str, float]:
             match = PREDICTION_SCORE_REGEXP.findall(x)[0]
             return {match[0]: float32(match[1])}
 
-        super().__init__(name, typefunc=typefunc, nafunc=lambda: dict(), **kwargs)
+        super().__init__(name, typefunc=typefunc, nafunc=lambda: {}, **kwargs)
 
 
 class DefaultAnnotationEntry(AnnotationEntry):
-    def __init__(self, name: str):
+    def __init__(self, name: str) -> None:
         super().__init__(name)
 
 
 class AnnotationTyper:
-    def __init__(self, mapping: Dict[str, AnnotationEntry]):
+    def __init__(self, mapping: dict[str, AnnotationEntry]) -> None:
         self._mapping = mapping
 
     def get_entry(self, key: str) -> AnnotationEntry:
@@ -336,14 +418,14 @@ class AnnotationTyper:
             entry = self._mapping[key]
         return entry
 
-    def convert(self, key: str, value: str) -> Tuple[str, AnnotationType]:
+    def convert(self, key: str, value: str) -> tuple[str, AnnotationType]:
         entry = self.get_entry(key)
         return entry.name, entry.convert(value)
 
 
 KNOWN_ANN_TYPE_MAP_SNPEFF = {
     "Allele": AnnotationEntry("Allele"),
-    "Annotation": AnnotationEntry("Annotation"),
+    "Annotation": ConsequenceEntry("Annotation"),
     "Annotation_Impact": AnnotationEntry("Annotation_Impact"),
     "Gene_Name": AnnotationEntry("Gene_Name"),
     "Gene_ID": AnnotationEntry("Gene_ID"),
@@ -353,9 +435,18 @@ KNOWN_ANN_TYPE_MAP_SNPEFF = {
     "Rank": AnnotationEntry("Rank"),
     "HGVS.c": AnnotationEntry("HGVS.c"),
     "HGVS.p": AnnotationEntry("HGVS.p"),
-    "cDNA.pos / cDNA.length": AnnotationEntry("cDNA", PosRange.from_snpeff_str),
-    "CDS.pos / CDS.length": AnnotationEntry("CDS", PosRange.from_snpeff_str),
-    "AA.pos / AA.length": AnnotationEntry("AA", PosRange.from_snpeff_str),
+    "cDNA.pos / cDNA.length": AnnotationEntry(
+        "cDNA.pos / cDNA.length",
+        PosRange.from_snpeff_str,
+    ),
+    "CDS.pos / CDS.length": AnnotationEntry(
+        "CDS.pos / CDS.length",
+        PosRange.from_snpeff_str,
+    ),
+    "AA.pos / AA.length": AnnotationEntry(
+        "AA.pos / AA.length",
+        PosRange.from_snpeff_str,
+    ),
     "Distance": AnnotationEntry("Distance", str),
     "ERRORS / WARNINGS / INFO": AnnotationListEntry(
         "ERRORS / WARNINGS / INFO",
@@ -371,7 +462,8 @@ KNOWN_ANN_TYPE_MAP_VEP = {
         description="In standard coordinate format (chr:start or chr:start-end)",
     ),
     "Allele": AnnotationEntry(
-        "Allele", description="The variant allele used to calculate the consequence"
+        "Allele",
+        description="The variant allele used to calculate the consequence",
     ),
     "Gene": AnnotationEntry("Gene", description="Ensembl stable ID of affected gene"),
     "Feature": AnnotationEntry("Feature", description="Ensembl stable ID of feature"),
@@ -380,24 +472,28 @@ KNOWN_ANN_TYPE_MAP_VEP = {
         description="Type of feature. "
         "Currently one of Transcript, RegulatoryFeature, MotifFeature.",
     ),
-    "Consequence": AnnotationEntry(
-        "Consequence", description="Consequence type of this variant"
+    "Consequence": ConsequenceEntry(
+        "Consequence",
+        description="Consequence type of this variant",
     ),
-    "cDNA_position": AnnotationEntry("cDNA", PosRange.from_vep_str),
-    "CDS_position": AnnotationEntry("CDS", PosRange.from_vep_str),
-    "Protein_position": AnnotationEntry("Protein", PosRange.from_vep_str),
+    "cDNA_position": AnnotationEntry("cDNA_position", PosRange.from_vep_str),
+    "CDS_position": AnnotationEntry("CDS_position", PosRange.from_vep_str),
+    "Protein_position": AnnotationEntry("Protein_position", PosRange.from_vep_str),
     "HGSVc": AnnotationEntry("HGSVc"),
     "HGSVp": AnnotationEntry("HGSVp"),
     "REF_ALLELE": AnnotationEntry("REF_ALLELE", description="The reference allele"),
     "IMPACT": AnnotationEntry(
-        "IMPACT", description="The impact modifier for the consequence type"
+        "IMPACT",
+        description="The impact modifier for the consequence type",
     ),
     "SYMBOL": AnnotationEntry("SYMBOL", description="The gene symbol"),
     "VARIANT_CLASS": AnnotationEntry(
-        "VARIANT_CLASS", description="Sequence Ontology variant class"
+        "VARIANT_CLASS",
+        description="Sequence Ontology variant class",
     ),
     "SYMBOL_SOURCE": AnnotationEntry(
-        "SYMBOL_SOURCE", description="The source of the gene symbol"
+        "SYMBOL_SOURCE",
+        description="The source of the gene symbol",
     ),
     "STRAND": AnnotationEntry(
         "STRAND",
@@ -405,7 +501,8 @@ KNOWN_ANN_TYPE_MAP_VEP = {
         description="The DNA strand (1 or -1) on which the transcript/feature lies",
     ),
     "ENSP": AnnotationEntry(
-        "ENSP", description="The Ensembl protein identifier of the affected transcript"
+        "ENSP",
+        description="The Ensembl protein identifier of the affected transcript",
     ),
     "FLAGS": AnnotationListEntry(
         "FLAGS",
@@ -418,10 +515,12 @@ KNOWN_ANN_TYPE_MAP_VEP = {
         description="Best match UniProtKB/Swiss-Prot accession of protein product",
     ),
     "TREMBL": AnnotationEntry(
-        "TREMBL", description="Best match UniProtKB/TrEMBL accession of protein product"
+        "TREMBL",
+        description="Best match UniProtKB/TrEMBL accession of protein product",
     ),
     "UNIPARC": AnnotationEntry(
-        "UNIPARC", description="Best match UniParc accession of protein product"
+        "UNIPARC",
+        description="Best match UniParc accession of protein product",
     ),
     "HGVSc": AnnotationEntry("HGVSc", description="The HGVS coding sequence name"),
     "HGVSp": AnnotationEntry("HGVSp", description="The HGVS protein sequence name"),
@@ -440,7 +539,8 @@ KNOWN_ANN_TYPE_MAP_VEP = {
         " with both given as prediction(score)",
     ),
     "PolyPhen": AnnotationPredictionScoreEntry(
-        "PolyPhen", description="The PolyPhen prediction and/or score"
+        "PolyPhen",
+        description="The PolyPhen prediction and/or score",
     ),
     "MOTIF_NAME": AnnotationEntry(
         "MOTIF_NAME",
@@ -476,20 +576,24 @@ KNOWN_ANN_TYPE_MAP_VEP = {
         "transcript for this gene",
     ),
     "CCDS": AnnotationEntry(
-        "CCDS", description="The CCDS identifer for this transcript, where applicable"
+        "CCDS",
+        description="The CCDS identifer for this transcript, where applicable",
     ),
-    "INTRON": AnnotationNumberTotalEntry(
+    "INTRON": AnnotationRangeTotalEntry(
         "INTRON", description="The intron number (out of total number)"
     ),
     "EXON": AnnotationRangeTotalEntry(
-        "EXON", description="The exon index range (out of total number of exons)"
+        "EXON",
+        description="The exon index range (out of total number of exons)",
     ),
     "DOMAINS": AnnotationListDictEntry(
         "DOMAINS",
         description="The source and identifer of any overlapping protein domains",
     ),
     "DISTANCE": AnnotationEntry(
-        "DISTANCE", int, description="Shortest distance from variant to transcript"
+        "DISTANCE",
+        int,
+        description="Shortest distance from variant to transcript",
     ),
     "IND": AnnotationEntry("IND", description="Individual name"),
     # "ZYG": AnnotationEntry(
@@ -500,7 +604,9 @@ KNOWN_ANN_TYPE_MAP_VEP = {
     #     "FREQS", description="Frequencies of overlapping variants used in filtering"
     # ),
     "AF": AnnotationEntry(
-        "AF", float32, description="Frequency of existing variant in 1000 Genomes"
+        "AF",
+        float32,
+        description="Frequency of existing variant in 1000 Genomes",
     ),
     "AFR_AF": AnnotationEntry(
         "AFR_AF",
@@ -583,8 +689,7 @@ KNOWN_ANN_TYPE_MAP_VEP = {
     "gnomAD_FIN_AF": AnnotationEntry(
         "gnomAD_FIN_AF",
         float32,
-        description="Frequency of existing variant in gnomAD exomes "
-        "Finnish population",
+        description="Frequency of existing variant in gnomAD exomes Finnish population",
     ),
     "gnomAD_NFE_AF": AnnotationEntry(
         "gnomAD_NFE_AF",
@@ -607,8 +712,7 @@ KNOWN_ANN_TYPE_MAP_VEP = {
     "MAX_AF": AnnotationEntry(
         "MAX_AF",
         float32,
-        description="Maximum observed allele frequency in "
-        "1000 Genomes, ESP and gnomAD",
+        description="Maximum observed allele frequency in 1000 Genomes, ESP and gnomAD",
     ),
     "MAX_AF_POPS": AnnotationListEntry(
         "MAX_AF_POPS",
@@ -621,7 +725,8 @@ KNOWN_ANN_TYPE_MAP_VEP = {
         description="ClinVar clinical significance of the dbSNP variant",
     ),
     "BIOTYPE": AnnotationEntry(
-        "BIOTYPE", description="Biotype of transcript or regulatory feature"
+        "BIOTYPE",
+        description="Biotype of transcript or regulatory feature",
     ),
     "APPRIS": AnnotationEntry(
         "APPRIS",
@@ -630,7 +735,8 @@ KNOWN_ANN_TYPE_MAP_VEP = {
         "NB: not available for GRCh37",
     ),
     "TSL": AnnotationEntry(
-        "TSL", description="Transcript support level. NB: not available for GRCh37"
+        "TSL",
+        description="Transcript support level. NB: not available for GRCh37",
     ),
     "PUBMED": AnnotationListEntry(
         "PUBMED",
@@ -682,10 +788,12 @@ KNOWN_ANN_TYPE_MAP_VEP = {
     #     "BAM_EDIT", description="Indicates success or failure of edit using BAM file"
     # ),
     "GIVEN_REF": AnnotationEntry(
-        "GIVEN_REF", description="Reference allele from input"
+        "GIVEN_REF",
+        description="Reference allele from input",
     ),
     "USED_REF": AnnotationEntry(
-        "USED_REF", description="Reference allele as used to get consequences"
+        "USED_REF",
+        description="Reference allele as used to get consequences",
     ),
     # TODO parse flag:
     #  see https://www.ensembl.org/info/docs/tools/vep/vep_formats.html#other_fields
@@ -713,18 +821,24 @@ KNOWN_ANN_TYPE_MAP_VEP = {
     #                 "the expected reference",
     # ),
     "AMBIGUITY": AnnotationEntry(
-        "AMBIGUITY", description="IUPAC allele ambiguity code"
+        "AMBIGUITY",
+        description="IUPAC allele ambiguity code",
     ),
     "Amino_acids": AnnotationListEntry(
-        "Amino_acids", description="Reference and variant amino acids", sep="/"
+        "Amino_acids",
+        description="Reference and variant amino acids",
+        sep="/",
     ),
     "Codons": AnnotationListEntry(
-        "Codons", description="Reference and variant codon sequence", sep="/"
+        "Codons",
+        description="Reference and variant codon sequence",
+        sep="/",
     ),
     # TODO HGNC_ID description
     "HGNC_ID": AnnotationEntry("HGNC_ID"),
     "MANE": AnnotationEntry(
-        "MANE", description="Matched Annotation from NCBI and EMBL-EBI (MANE)."
+        "MANE",
+        description="Matched Annotation from NCBI and EMBL-EBI (MANE).",
     ),
     "MANE_SELECT": AnnotationEntry(
         "MANE_SELECT",
@@ -769,12 +883,74 @@ KNOWN_ANN_TYPE_MAP_VEP = {
     "ExACpLI": AnnotationEntry(
         "ExACpLI",
         typefunc=float32,
-        description="Probabililty of a gene being loss-of-function intolerant (pLI). "
+        description="Probability of a gene being loss-of-function intolerant (pLI). "
         "Please cite the respective ExAC publication alongside the VEP if you use this "
         "resource: https://www.ncbi.nlm.nih.gov/pubmed/27535533",
     ),
+    "am_class": AnnotationEntry(
+        "am_class",
+        description="AlphaMissense classification of variant "
+        "into one of three discrete categories: "
+        "'Likely pathogenic', 'Likely benign', or 'ambiguous'"
+        "Please cite the AlphaMissense publication alongside Ensembl VEP "
+        "if you use this resource: https://doi.org/10.1126/science.adg7492",
+    ),
+    "am_pathogenicity": AnnotationEntry(
+        "am_pathogenicity",
+        typefunc=float32,
+        description="AlphaMissense score of a variant being "
+        "likely pathogenic, likely benign, or ambiguous."
+        "Please cite the AlphaMissense publication alongside Ensembl VEP "
+        "if you use this resource: https://doi.org/10.1126/science.adg7492",
+    ),
+    "SpliceAI_pred_DS_AG": AnnotationEntry(
+        "SpliceAI_pred_DS_AG",
+        typefunc=float32,
+        description="Probability of variant being splice-altering acceptor gain. "
+        "Please cite the respective ExAC publication alongside the VEP if you use this "
+        "resource: https://www.ncbi.nlm.nih.gov/pubmed/30661751",
+    ),
+    "SpliceAI_pred_DS_AL": AnnotationEntry(
+        "SpliceAI_pred_DS_AL",
+        typefunc=float32,
+        description="Probability of variant being splice-altering acceptor loss. "
+        "Please cite the respective ExAC publication alongside the VEP if you use this "
+        "resource: https://www.ncbi.nlm.nih.gov/pubmed/30661751",
+    ),
+    "SpliceAI_pred_DS_DG": AnnotationEntry(
+        "SpliceAI_pred_DS_DG",
+        typefunc=float32,
+        description="Probability of variant being splice-altering donor gain. "
+        "Please cite the respective ExAC publication alongside the VEP if you use this "
+        "resource: https://www.ncbi.nlm.nih.gov/pubmed/30661751",
+    ),
+    "SpliceAI_pred_DS_DL": AnnotationEntry(
+        "SpliceAI_pred_DS_DL",
+        typefunc=float32,
+        description="Probability of variant being splice-altering donor loss. "
+        "Please cite the respective ExAC publication alongside the VEP if you use this "
+        "resource: https://www.ncbi.nlm.nih.gov/pubmed/30661751",
+    ),
 }
 
+# VEP now uses gnomADe_* and gnomADg_* instead of gnomAD_*.
+# Therefore, we need to add these new entries to the KNOWN_ANN_TYPE_MAP_VEP
+NEW_GNOMAD_ENTRIES = {}
+for key, typer in KNOWN_ANN_TYPE_MAP_VEP.items():
+    if key.startswith("gnomAD_"):
+        sub_key = key.split("_", 1)[1]
+        for t in ("e", "g"):
+            name = f"gnomAD{t}_{sub_key}"
+            description = typer.description()
+            if t == "g":
+                description = description.replace("gnomAD exomes", "gnomAD genomes")
+            entry = AnnotationEntry(
+                name=name,
+                typefunc=typer._typefunc,
+                description=description,
+            )
+            NEW_GNOMAD_ENTRIES[name] = entry
+KNOWN_ANN_TYPE_MAP_VEP.update(NEW_GNOMAD_ENTRIES)
 KNOWN_ANN_TYPE_MAP = {**KNOWN_ANN_TYPE_MAP_SNPEFF, **KNOWN_ANN_TYPE_MAP_VEP}
 
 ANN_TYPER = AnnotationTyper(KNOWN_ANN_TYPE_MAP)
